import numpy as np
from dataclasses import asdict
from cratonapi.dataconnector import DataConnector
from cratonapi import datacontainers


class Cube:
    """
    Класс для работы с сейсмическим кубом.

    Attributes
    ----------
    connection : cratonapi.dataconnector.DataConnector
        Объект класса для взаимодействия с WSeis.
    """

    def __init__(self, connection: DataConnector):
        self.connection = connection

    def read(self,
             cube_id: int,
             horizon_top_name: str,
             horizon_bot_name: str,
             top_off: float,
             bot_off: float,
             start_inline_idx: int,
             end_inline_idx: int,
             start_crossline_idx: int,
             end_crossline_idx: int) -> datacontainers.CubeDataSlice:
        """
        Подгружает данные трасс куба в диапазоне.

        Parameters
        ----------
        cube_id: int
            Идентификатор куба.
        horizon_top_name: str
            Имя верхнего горизонта границы.
        horizon_bot_name: str
            Имя нижнего горизонта границы.
        top_off: float
            Смещение для верхнего горизонта.
        bot_off: float
            Смещение для нижней горизонта.
        start_inline_idx: int
            Индекс инлайна с которого начинать набор данных.
        end_inline_idx: int
            Индекс инлайна которым заканчивать набор данных.
        start_crossline_idx: int
            Индекс кросслайна с которого начинать набор данных.
        end_crossline_idx: int
            Индекс кросслайна которым заканчивать набор данных.

        Returns
        -------
        cratonapi.datacontainers.CubeDataSlice
            Куб.
        """

        cube = self.connection.get_cube_data_range(cube_id=cube_id,
                                                   horizon_top_name=horizon_top_name,
                                                   horizon_bot_name=horizon_bot_name,
                                                   top_off=top_off,
                                                   bot_off=bot_off,
                                                   start_inline_idx=start_inline_idx,
                                                   end_inline_idx=end_inline_idx,
                                                   start_crossline_idx=start_crossline_idx,
                                                   end_crossline_idx=end_crossline_idx)
        return cube

    def get_info(self) -> dict:
        """
        Подгружает информацию о кубах.

        Returns
        -------
        dict
            Словарь, где ключ - имя куба.
            Значения:
                id: int
                    Идентификатор куба.
                x_min_inl_min_xl: float
                    X координата трассы с минимальными инлайном и кросслайном в кубе.
                y_min_inl_min_xl: float
                    Y координата трассы с минимальными инлайном и кросслайном в кубе.
                x_max_inl_min_xl: float
                    X координата трассы с максимальным инлайном и минимальным кросслайном в кубе.
                y_max_inl_min_xl: float
                    Y координата трассы с максимальным инлайном и минимальным кросслайном в кубе.
                x_min_inl_max_xl: float
                    X координата трассы с минимальным инлайном и максимальным кросслайном в кубе.
                y_min_inl_max_xl: float
                    Y координата трассы с минимальным инлайном и максимальным кросслайном в кубе.
                x_max_inl_max_xl: float
                    X координата трассы с максимальными инлайном и кросслайном в кубе.
                y_max_inl_max_xl: float
                    Y координата трассы с максимальными инлайном и кросслайном в кубе.
                inline_count: int
                    Количество инлайнов в кубе.
                xline_count: int
                    Количество кросслайнов в кубе.
                samples_count: int
                    Количество отсчётов в каждой трассе куба.
                dt: int
                    Шаг дискретизации трасс куба.
                cube_type: int
                    Тип куба.
                min_idx_inline: int
                    Минимальный номер инлайна.
                min_idx_xline: int
                    Минимальный номер кросслайна.
                horizons: list[str]
                    Список имен кубов.
                transform_matrix: cratonapi.datacontainers.TransformMatrix
                    Матрица трансформации.
        """

        cube_list = self.connection.get_cubes()
        info = {}
        for cube in cube_list:
            id_dict = {'id': cube.cube_id}
            properties_dict = asdict(self.connection.get_cube_properties(cube.cube_id))
            horizons_dict = {'horizons': list(self.connection.get_cube_horizons(cube.cube_id))}
            transforms_dict = {'transform_matrix': self.connection.get_transform_matrix(cube.cube_id)}

            info_dict = dict(id_dict, **properties_dict, **horizons_dict, **transforms_dict)
            info[cube.cube_name] = info_dict
        return info

    @staticmethod
    def __get_distance(coord_1, coord_2):
        return np.sqrt((coord_1[0] - coord_2[0]) ** 2 + (coord_1[1] - coord_2[1]) ** 2)

    @staticmethod
    def __get_transform(transform_matrix, x, y):
        x_ = (transform_matrix.a * x + transform_matrix.b * y + transform_matrix.c) / (
                transform_matrix.g * x + transform_matrix.h * y + 1)
        y_ = (transform_matrix.d * x + transform_matrix.e * y + transform_matrix.f) / (
                transform_matrix.g * x + transform_matrix.h * y + 1)
        return x_, y_

    @staticmethod
    def parse_to_numpy(cube: datacontainers.CubeDataSlice,
                       cube_props: dict) -> tuple[np.ndarray, np.ndarray, np.ndarray, float]:
        """
        Преобразовывает куб в numpy массив.

        Parameters
        ----------
        cube: cratonapi.datacontainers.CubeDataSlice
            Куб.
        cube_props: dict
            Словарь со значениями:
                x_min_inl_min_xl: float
                    X координата трассы с минимальными инлайном и кросслайном в кубе.
                y_min_inl_min_xl: float
                    Y координата трассы с минимальными инлайном и кросслайном в кубе.
                x_max_inl_min_xl: float
                    X координата трассы с максимальным инлайном и минимальным кросслайном в кубе.
                y_max_inl_min_xl: float
                    Y координата трассы с максимальным инлайном и минимальным кросслайном в кубе.
                x_min_inl_max_xl: float
                    X координата трассы с минимальным инлайном и максимальным кросслайном в кубе.
                y_min_inl_max_xl: float
                    Y координата трассы с минимальным инлайном и максимальным кросслайном в кубе.
                x_max_inl_max_xl: float
                    X координата трассы с максимальными инлайном и кросслайном в кубе.
                y_max_inl_max_xl: float
                    Y координата трассы с максимальными инлайном и кросслайном в кубе.
                inline_count: int
                    Количество инлайнов в кубе.
                xline_count: int
                    Количество кросслайнов в кубе.
                transform_matrix: cratonapi.datacontainers.TransformMatrix
                    Матрица трансформации.

        Returns
        -------
        tuple[np.ndarray, np.ndarray, np.ndarray, float]
            2D массив трасс куба.
            Сетка координат по x в виде 2D массива.
            Сетка координат по y в виде 2D массива.
            Код бланковки.
        """

        inlines = cube.inlines
        xlines = cube.xlines

        x_cube_coords = np.array([cube_props['x_min_inl_min_xl'],
                                  cube_props['x_max_inl_min_xl'],
                                  cube_props['x_min_inl_max_xl'],
                                  cube_props['x_max_inl_max_xl'], ])

        y_cube_coords = np.array([cube_props['y_min_inl_min_xl'],
                                  cube_props['y_max_inl_min_xl'],
                                  cube_props['y_min_inl_max_xl'],
                                  cube_props['y_max_inl_max_xl'], ])

        data = cube.data
        blank_code = np.max(abs(data))
        blank_mask = data == blank_code

        cube_data = [data[i, ~blank_mask[i, :]] for i in range(len(blank_mask))]
        shapes = np.array([cube_data[i].shape[0] for i in range(len(cube_data))])
        min_size = np.min(shapes)
        if min_size == 0:
            idx = np.where(shapes == 0)[0]
            for i in idx:
                if i > 0:
                    cube_data[i] = cube_data[i - 1]
                else:
                    cube_data[i] = cube_data[i + 1]
                shapes[i] = cube_data[i].shape[0]
            min_size = np.min(shapes)
        cube_slice = np.stack([cube_data[i][:min_size] for i in range(len(cube_data))])

        dx = Cube.__get_distance([x_cube_coords[0], y_cube_coords[0]],
                                 [x_cube_coords[1], y_cube_coords[1]]) / cube_props['inline_count']
        dy = Cube.__get_distance([x_cube_coords[0], y_cube_coords[0]],
                                 [x_cube_coords[2], y_cube_coords[2]]) / cube_props['xline_count']

        dx, dy = int(dx), int(dy)

        idx_dict = {(int(inlines[i]), int(xlines[i])): i for i in range(inlines.shape[0])}

        x_min = np.min(x_cube_coords)
        y_min = np.min(y_cube_coords)

        x_max = np.max(x_cube_coords)
        y_max = np.max(y_cube_coords)

        size = (int((x_max - x_min) / dx), int((y_max - y_min) / dy))
        grid_values = np.zeros((size[1], size[0], len(cube_slice[0])))
        for i in range(size[1]):
            y_i = y_min + i * dy
            for j in range(size[0]):
                x_j = x_min + j * dx
                inl, xl = Cube.__get_transform(cube_props['transform_matrix'], x_j, y_i)
                inl = int(inl * cube_props['inline_count'])
                xl = int(xl * cube_props['xline_count'])
                if (inl, xl) in idx_dict:
                    idx = idx_dict[(inl, xl)]
                    grid_values[i, j, :] = cube_slice[idx][:]
                else:
                    grid_values[i, j, :] = np.nan
        grid_values = np.asarray(grid_values)
        grid_values = np.reshape(grid_values,
                                 newshape=(grid_values.shape[0] * grid_values.shape[1],
                                           grid_values.shape[2]))

        x = np.linspace(x_min, x_max, size[0])
        y = np.linspace(y_min, y_max, size[1])

        xx, yy = np.meshgrid(x, y)

        return grid_values, xx, yy, blank_code



import numpy as np
from cratonapi.dataconnector import DataConnector
from scipy.interpolate import RegularGridInterpolator
from cratonapi import datacontainers


class Grid:
    """
    Класс для работы с гридами.

    Attributes
    ----------
    connection : cratonapi.dataconnector.DataConnector
        Объект класса для взаимодействия с WSeis.
    """

    def __init__(self, connection: DataConnector):
        self.connection = connection

    def read(self, grid_id_list: list[int], grid_type_list: list[int]) -> list[datacontainers.Grid]:
        """
        Подгружает cетки(гриды).

        Parameters
        ----------
        grid_id_list: list[int]
            Список идентификаторов гридов.
        grid_type_list: list[int]
            Список типов гридов. Тип грида: 0 — обычный, 1 — смешанный.

        Returns
        -------
        list[cratonapi.datacontainers.Grid]
            Список гридов.
        """

        grid_list = []
        for grid_id, grid_type in zip(grid_id_list, grid_type_list):
            if grid_type == 0:
                grid = self.connection.get_grid(grid_id)
                if grid.data.shape[0] != 0:
                    grid_list.append(grid)
            elif grid_type == 1:
                grids = self.connection.get_grid_sp(grid_id)
                for grid in grids:
                    grid_list.append(grid)
        return grid_list

    def save(self, name: str, grid: datacontainers.Grid) -> None:
        """
        Загружает грид в Desmana.

        Parameters
        ----------
        name: str
            Имя грида.
        grid: cratonapi.datacontainers.Grid
            Сеточная модель.
        """

        self.connection.upload_grid(name, grid)

    def get_info(self) -> dict:
        """
        Подгружает информацию о гридах.

        Returns
        -------
        dict
            Словарь, где ключ - имя грида.
            Значения:
                id: int
                    Идентификатор грида.
                type: int
                    Тип грида: 0 — обычный, 1 — смешанный.
                visibility: int
                    Видимость грида: 0 — скрытый на планшете, 1 — отображается на планшете.
        """

        grids_list = self.connection.get_grids_list()
        info = {grid.grid_name: {'id': grid.grid_id,
                                 'type': grid.grid_type,
                                 'visibility': grid.grid_visibility}
                for grid in grids_list}
        return info

    @staticmethod
    def numpy_to_grid(data: np.ndarray,
                      x: np.ndarray,
                      y: np.ndarray,
                      blank_code: float,
                      grid_id: int = 0) -> datacontainers.Grid:
        """
        Преобразовывает numpy массив в грид.

        Parameters
        ----------
        data: np.ndarray
            1D массив значений грида.
        x: np.ndarray
            Сетка координат по x в виде 2D массива.
        y: np.ndarray
            Сетка координат по y в виде 2D массива.
        blank_code: float
            Код бланковки.
        grid_id: int, default=0
            Идентификатор грида.

        Returns
        -------
        cratonapi.datacontainers.Grid
            Грид.
        """

        assert data.ndim == 1, "data: Ожидался 1D массив"
        assert x.ndim == 2, "x: Ожидался 2D массив"
        assert y.ndim == 2, "y: Ожидался 2D массив"
        assert grid_id >= 0, "grid_id должен принимать неотрицательное значение"

        data_to_save = np.nan_to_num(data, nan=blank_code)
        grid_to_save = datacontainers.Grid(n_id=grid_id,
                                           n_x=x.shape[1],
                                           n_y=x.shape[0],
                                           x_min=np.min(x),
                                           x_max=np.max(x),
                                           y_min=np.min(y),
                                           y_max=np.max(y),
                                           z_min=np.nanmin(data[data != blank_code]),
                                           z_max=np.nanmax(data[data != blank_code]),
                                           blank_code=blank_code,
                                           data=data_to_save)
        return grid_to_save

    @staticmethod
    def parse_to_numpy(grid_list: list[datacontainers.Grid]) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
        """
        Преобразовывает список гридов в numpy массив.

        Parameters
        ----------
        grid_list: list[cratonapi.datacontainers.Grid]
            Список гридов.

        Returns
        -------
        tuple[np.ndarray, np.ndarray, np.ndarray]
            2D массив значений грида.
            Сетка координат по x в виде 2D массива.
            Сетка координат по y в виде 2D массива.
        """

        x_min = np.max([grid.x_min for grid in grid_list])
        x_max = np.min([grid.x_max for grid in grid_list])

        y_min = np.max([grid.y_min for grid in grid_list])
        y_max = np.min([grid.y_max for grid in grid_list])

        dx = np.min([(grid.x_max - grid.x_min) / grid.n_x for grid in grid_list])
        dy = np.min([(grid.y_max - grid.y_min) / grid.n_y for grid in grid_list])

        x = np.arange(x_min, x_max, dx)
        y = np.arange(y_min, y_max, dy)

        xx, yy = np.meshgrid(x, y)
        grid_data_list = []
        mask_list = []
        for grid in grid_list:
            blank_code = grid.blank_code
            x_grid = np.linspace(grid.x_min, grid.x_max, grid.n_x)
            y_grid = np.linspace(grid.y_min, grid.y_max, grid.n_y)
            grid_2d = np.reshape(grid.data, (grid.n_y, grid.n_x))
            grid_2d[grid_2d == blank_code] = np.nan

            interp_func = RegularGridInterpolator((y_grid, x_grid), grid_2d)
            new_grid = interp_func((yy, xx))
            mask = np.isnan(new_grid)
            mask_list.append(mask.ravel())
            grid_data_list.append(new_grid.ravel())

        mask_list = np.asarray(mask_list)
        general_mask = np.logical_or.reduce(mask_list).T
        data = np.asarray(grid_data_list).T
        data[general_mask] = np.nan
        return data, xx, yy



import numpy as np
from scipy.signal import find_peaks


def remove_single_width_peaks(data, peaks):
    width = 1
    filled_data = np.copy(data)
    for i in peaks:
        left = max(0, i - width)
        right = min(len(data) - 1, i + width + 1)
        filled_data[i] = np.median(filled_data[left:right])
    return filled_data.astype('int')


def remove_given_width_peaks(data, peaks, width):
    window_size = width + width // 2
    filled_data = np.copy(data)
    for i in peaks:
        peak_left = max(0, i - width)
        peak_right = min(len(data) - 1, i + width + 1)
        left = max(0, i - window_size)
        right = min(len(data) - 1, i + window_size + 1)
        filled_data[peak_left:peak_right] = np.median(filled_data[left:right])
    return filled_data.astype('int')


def get_peaks(curve, width):
    positive_peaks, _ = find_peaks(curve, width=[0, width])
    negative_peaks, _ = find_peaks(abs(curve - 1), width=[0, width])
    peaks = np.unique(np.hstack((positive_peaks, negative_peaks)))
    return peaks



import numpy as np
from cratonapi.dataconnector import DataConnector
from cratonapi import datacontainers
from scipy.interpolate import interp1d


class Well:
    """
    Класс для работы со скважинами.

    Attributes
    ----------
    connection : cratonapi.dataconnector.DataConnector
        Объект класса для взаимодействия с WSeis.
    """

    def __init__(self, connection: DataConnector):
        self.connection = connection

    def read(self, well_id: int, well_curves_name: list[str]) -> list[datacontainers.WellCurve]:
        """
        Подгружает кривые из скважины.

        Parameters
        ----------
        well_id: int
            Идентификатор скважины.
        well_curves_name: list[str]
            Список имен кривых.

        Returns
        -------
        list[cratonapi.datacontainers.WellCurve]
            Список кривых для скважины.
        """

        assert well_id >= 0, "well_id должен принимать неотрицательное значение"

        well_curves = self.connection.get_well_curves(well_id)
        curves = []
        for curve in well_curves:
            if curve.curve_name in well_curves_name:
                curves.append(curve)
        return curves

    @staticmethod
    def parse_to_numpy(well_curves_list: list[datacontainers.WellCurve]) -> tuple[np.ndarray, np.ndarray]:
        """
        Преобразовывает список кривых в numpy массив.

        Parameters
        ----------
        well_curves_list: list[cratonapi.datacontainers.WellCurve]
            Список кривых, длины N

        Returns
        -------
        tuple[np.ndarray, np.ndarray]
            Массив значений кривых, размером (M, N).
            Массив значений глубин для всех кривых, размером (M, ).
        """

        depths = []
        values = []
        for well_curve in well_curves_list:
            values.append(well_curve.point_values)
            depths.append(well_curve.point_depths)

        min_depth = np.min([np.min(depths[i]) for i in range(len(depths))])
        max_depth = np.max([np.max(depths[i]) for i in range(len(depths))])

        dh = np.min([np.min(np.abs(np.diff(depths[i]))) for i in range(len(depths))])
        dh = np.round(dh, 2)

        values_new = []
        full_depth = np.arange(min_depth, max_depth, dh)
        for i in range(len(depths)):
            mask = ~np.isnan(values[i])
            depths_i = depths[i][mask]
            values_i = values[i][mask]
            depths_i_interpolated = np.arange(np.min(depths_i), np.max(depths_i), dh)
            depths_i_interpolated = depths_i_interpolated[depths_i_interpolated < np.max(depths_i)]
            interp_func = interp1d(depths_i, values_i, kind='linear')
            values_i_interpolated = interp_func(depths_i_interpolated)
            idx_min = int((depths_i_interpolated[0] - full_depth[0]) // dh)
            idx_max = int((depths_i_interpolated[-1] - full_depth[0]) // dh)
            values_i_new = np.zeros(len(full_depth))
            values_i_new[:idx_min] = np.nan
            values_i_new[idx_max:] = np.nan
            values_i_new[idx_min: idx_min + len(values_i_interpolated)] = values_i_interpolated
            values_new.append(values_i_new)
        values_new = np.stack(values_new).T
        return values_new, full_depth

    @staticmethod
    def numpy_to_curve(values: np.ndarray,
                       depths: np.ndarray,
                       curve_name: str,
                       curve_type: int = 0) -> datacontainers.WellCurve:
        """
        Преобразовывает numpy массив в кривую.

        Parameters
        ----------
        values: np.ndarray
            1D массив значений кривой, размером N.
        depths: np.ndarray
            1D массив значений глубин кривой, размером N.
        curve_name: str
            Имя кривой.
        curve_type: int, default=0
            Тип кривой.

        Returns
        -------
        cratonapi.datacontainers.WellCurve
            Кривая.
        """

        assert values.ndim == 1, "values: Ожидался 1D массив"
        assert depths.ndim == 1, "depths: Ожидался 1D массив"
        assert values.shape == depths.shape, "Размеры values и depths должны быть равны"
        assert curve_type >= 0, "curve_type должен принимать неотрицательное значение"

        curve_to_save = datacontainers.WellCurve(curve_type=curve_type,
                                                 curve_name=curve_name,
                                                 point_values=values,
                                                 point_depths=depths)
        return curve_to_save

    def save(self, well_curve: datacontainers.WellCurve, well_id: int) -> None:
        """
        Загружает кривую в GISWell.

        Parameters
        ----------
        well_curve: cratonapi.datacontainers.WellCurve
            Кривая.
        well_id: int
            Идентификатор скважины.
        """

        assert well_id >= 0, "well_id должен принимать неотрицательное значение"

        self.connection.upload_well_curves(well_curve, well_id)

    def get_info(self) -> dict:
        """
        Подгружает информацию о скважинах.

        Returns
        -------
        dict
            Словарь, где ключ - имя скважины.
            Значения:
                id: int
                    идентификатор скважины.
                curves_name: list[str]
                    Список имен кривых в скважине.
                strat_levels: dict
                    Словарь, где ключ - имя стратиграфического уровня.
                        Значения:
                            'level_id': int
                                Идентификатор стратиграфического уровня.
                            'level_depth': int
                                Глубина стратиграфического уровня.
                            'level_age': float
                                Возраст уровня (млн.лет).
        """

        well_name = self.connection.get_wells_list()
        strat_levels_dict = self.get_all_strat_levels_dict()

        info = {}
        for well in well_name:
            id_dict = {'id': well.well_id}
            well_curves = self.connection.get_well_curves(well.well_id)

            curves_name_dict = {'curves_name': [curve.curve_name for curve in well_curves]}

            well_levels_depth_dict = self.get_well_levels_depths(well.well_id)
            well_strat_info = Well.__parse_well_dicts(well_levels_depth_dict, strat_levels_dict)
            well_strat_dict = {'strat_levels': well_strat_info}

            info_dict = dict(**id_dict, **curves_name_dict, **well_strat_dict)
            info[well.well_name] = info_dict
        return info

    def get_all_strat_levels_dict(self) -> dict:
        """
        Подгружает информацию о всех стратиграфических уровнях.

        Returns
        -------
        dict
            Словарь, где ключ - идентификатор стратиграфического уровня.
            Значения:
                'level_age': float
                    Возраст уровня (млн.лет).
                'level_name': str
                    Имя стратиграфического уровня.
        """

        strat_levels = self.connection.get_stratigraphic_levels()
        strat_levels_dict = {}
        for i in range(len(strat_levels)):
            strat_levels_dict[strat_levels[i].level_id] = {'level_age': strat_levels[i].level_age,
                                                           'level_name': strat_levels[i].level_name}
        return strat_levels_dict

    def get_well_levels_depths(self, well_id: int) -> dict:
        """
        Подгружает глубины стратиграфических уровней в скважине.

        Parameters
        ----------
        well_id: int
            Идентификатор скважины.

        Returns
        -------
        dict
            Словарь, где ключ - идентификатор стратиграфического уровня.
            Значения:
                'level_depth': int
                    Глубина стратиграфического уровня.
        """

        assert well_id >= 0, "well_id должен принимать неотрицательное значение"

        well_levels_depth = self.connection.get_well_stratigraphic_levels(well_id)
        well_levels_depth_dict = {}
        for i in range(len(well_levels_depth)):
            well_levels_depth_dict[well_levels_depth[i].level_id] = {'level_depth': well_levels_depth[i].level_depth}
        return well_levels_depth_dict

    @staticmethod
    def __parse_well_dicts(well_levels_depth_dict, strat_levels_dict):
        well_strat_info = {}
        for level_id in well_levels_depth_dict.keys():
            level_name = strat_levels_dict[level_id]['level_name']
            well_strat_info[level_name] = {'level_id': level_id,
                                           'level_depth': well_levels_depth_dict[level_id]['level_depth'],
                                           'level_age': strat_levels_dict[level_id]['level_age']}
        return well_strat_info

    @staticmethod
    def get_min_max_depth(well_curves_list: list[datacontainers.WellCurve]) -> tuple[float, float]:
        """
        Считает минимальную и максимальную глубину среди всех кривых.

        Parameters
        ----------
        well_curves_list: list[cratonapi.datacontainers.WellCurve]
            Список кривых.

        Returns
        -------
        tuple[float, float]
            Минимальная глубина среди всех кривых.
            Максимальная глубина среди всех кривых.
        """

        depths = [curve.point_depths for curve in well_curves_list]
        min_depth = np.min([np.min(depths[i]) for i in range(len(depths))])
        max_depth = np.max([np.max(depths[i]) for i in range(len(depths))])
        return min_depth, max_depth



from typing import Tuple, Any

from numpy import ndarray, dtype
from sklearn.cluster import KMeans, HDBSCAN
import numpy as np
from scipy import stats
from sklearn.mixture import GaussianMixture

RANDOM_STATE = 42
ALPHA_K = 0.02


class KMeansClassifier:
    """Класс для KMeans кластеризации."""

    @staticmethod
    def __get_scaled_inertia(data, cluster):
        inertia_o = np.square((data - data.mean(axis=0))).sum()
        scaled_inertia = cluster.inertia_ / inertia_o + ALPHA_K * cluster.n_clusters
        return scaled_inertia

    @staticmethod
    def get_cluster(data, n_clusters):
        cluster = KMeans(n_clusters=n_clusters,
                         random_state=RANDOM_STATE, algorithm='elkan')
        cluster.fit(data)
        return cluster

    @staticmethod
    def __get_cluster_centers(kmeans):
        kmeans_centers = np.unique(np.sum(kmeans.cluster_centers_, axis=1))
        return kmeans_centers

    @staticmethod
    def __get_labels(kmeans):
        kmeans_labels_before_sorting = kmeans.labels_
        kmeans_centers = KMeansClassifier.__get_cluster_centers(kmeans)
        kmeans_uniq_labels = np.unique(kmeans_labels_before_sorting)
        sorted_kmeans_uniq_labels = kmeans_uniq_labels[np.argsort(kmeans_centers)]

        kmeans_labels = np.zeros_like(kmeans_labels_before_sorting)
        for i, label in enumerate(sorted_kmeans_uniq_labels):
            kmeans_labels[kmeans_labels_before_sorting == label] = i

        return kmeans_labels


    @staticmethod
    def find_the_best_number_of_clusters(data: np.ndarray) -> int:
        """
        Подбирает оптимальное количество кластеров на основании взвешенной инерции.

        Parameters
        ----------
        data: np.ndarray
            2D массив значений.

        Returns
        -------
        int
            Оптимальное количество кластеров.
        """

        assert data.ndim == 2, "data: Ожидался 2D массив"

        scaled_inertia_list = []
        number_of_clusters_list = range(2, min(15, data.shape[0]))
        for number_of_clusters in number_of_clusters_list:
            cluster = KMeansClassifier.get_cluster(data, number_of_clusters)
            scaled_inertia = KMeansClassifier.__get_scaled_inertia(data, cluster)
            scaled_inertia_list.append(scaled_inertia)
        scaled_inertia_list = np.asarray(scaled_inertia_list)
        best_number_of_clusters = number_of_clusters_list[np.argmin(scaled_inertia_list)]
        return best_number_of_clusters

    @staticmethod
    def calculate(data: np.ndarray, n_clusters: int) -> tuple[KMeans, ndarray[Any, dtype[Any]] | Any]:
        """
        Распределяет объекты по n_clusters кластерам по сходству. Для вычисления этого сходства используется евклидово расстояние в качестве меры.

        Parameters
        ----------
        data: np.ndarray
            Массив значений (N, M). Где N-количество точек, M - количество атрибутов.
        n_clusters: int
            Количество кластеров.

        Returns
        -------
        np.ndarray
            Массив меток размера N.
        """

        assert data.ndim == 2, "data: Ожидался 2D массив"
        assert n_clusters > 0, "Параметр n_clusters должен принимать положительное значение"
        cluster = KMeansClassifier.get_cluster(data, n_clusters)
        labels = KMeansClassifier.__get_labels(cluster)
        return cluster, labels


class GaussianMixtureClassifier:
    """Класс для Gaussian Mixture кластеризации."""

    @staticmethod
    def get_cluster(data, n_clusters):
        cluster = GaussianMixture(n_components=n_clusters,
                                  random_state=RANDOM_STATE)
        cluster.fit(data)
        return cluster

    @staticmethod
    def __get_cluster_centers(gm, data):
        centers = np.empty(shape=(gm.n_components, data.shape[1]))
        for i in range(gm.n_components):
            density = stats.multivariate_normal(cov=gm.covariances_[i], mean=gm.means_[i]).logpdf(data)
            centers[i, :] = data[np.argmax(density)]
        centers = np.sum(centers, axis=1)
        return centers

    @staticmethod
    def __get_labels(gm, data):
        gm_labels_before_sorting = gm.predict(data)
        gm_centers = GaussianMixtureClassifier.__get_cluster_centers(gm, data)
        gm_uniq_labels = np.unique(gm_labels_before_sorting)
        sorted_gm_uniq_labels = gm_uniq_labels[np.argsort(gm_centers)]

        gm_labels = np.zeros_like(gm_labels_before_sorting)
        for i, label in enumerate(sorted_gm_uniq_labels):
            gm_labels[gm_labels_before_sorting == label] = i

        return gm_labels

    @staticmethod
    def find_the_best_number_of_clusters(data: np.ndarray) -> int:
        """
        Подбирает оптимальное количество кластеров на основе Байесовского критерия.

        Parameters
        ----------
        data: np.ndarray
            2D массив значений.

        Returns
        -------
        int
            Оптимальное количество кластеров.
        """

        assert data.ndim == 2, "data: Ожидался 2D массив"

        bic_list = []
        number_of_clusters_list = range(2, min(15, data.shape[0] + 1))
        for number_of_clusters in number_of_clusters_list:
            gm = GaussianMixtureClassifier.get_cluster(data, number_of_clusters)
            bic = -gm.bic(data)
            if bic < 0:
                bic = np.nan
            bic_list.append(bic)
        bic_list = np.asarray(bic_list)
        # bic_list_log = np.log(bic_list)
        # best_number_of_clusters = number_of_clusters_list[np.nanargmin(bic_list_log)]
        best_number_of_clusters = number_of_clusters_list[np.nanargmin(bic_list)]
        return best_number_of_clusters

    @staticmethod
    def calculate(data: np.ndarray, n_clusters: int) -> tuple[GaussianMixture, ndarray[Any, dtype[Any]] | Any]:
        """
        Распределяет объекты по n_clusters кластерам, предполагая, что данные состоят из смеси гауссовых распределений.

        Parameters
        ----------
        data: np.ndarray
            Массив значений (N, M). Где N - количество точек, M - количество атрибутов.
        n_clusters: int
            Количество кластеров.

        Returns
        -------
        np.ndarray
            Массив меток размера N.
        """

        assert data.ndim == 2, "data: Ожидался 2D массив"
        assert n_clusters > 0, "Параметр n_clusters должен принимать положительное значение"
        cluster = GaussianMixtureClassifier.get_cluster(data=data, n_clusters=n_clusters)
        labels = GaussianMixtureClassifier.__get_labels(gm=cluster, data=data)
        return cluster, labels


class HDBSCANClassifier:
    """Класс для HDBSCAN кластеризации."""

    @staticmethod
    def get_cluster(data, min_cluster_size, min_samples, cluster_selection_epsilon):
        cluster = HDBSCAN(min_cluster_size=min_cluster_size, min_samples=min_samples,
                          cluster_selection_epsilon=cluster_selection_epsilon)
        cluster.fit(data)
        return cluster

    @staticmethod
    def __get_labels(hdbscan):
        hdbscan_labels = hdbscan.labels_
        return hdbscan_labels

    @staticmethod
    def get_number_of_clusters(labels: np.ndarray) -> int:
        """
        Возвращает количество кластеров.

        Parameters
        ----------
        labels: np.ndarray
            1D массив меток.

        Returns
        -------
        int
            Количество кластеров.
        """

        assert labels.ndim == 1, "labels: Ожидался 1D массив"

        return np.unique(labels).shape[0]

    @staticmethod
    def calculate(data: np.ndarray,
                  min_cluster_size: int = 5,
                  min_samples: int = None,
                  cluster_selection_epsilon: float = 0.0) -> tuple[HDBSCAN, Any]:
        """
        Выявляет кластеры в наборе данных на основе распределения плотности точек данных.

        Parameters
        ----------
        data: np.ndarray
            Массив значений (N, M). Где N-количество точек, M - количество атрибутов.
        min_cluster_size: int, default=5
            Минимальное количество выборок, чтобы эта группа считалась кластером(параметр HDBSCAN).
        min_samples: int, default=None
            Минимальное количество выборок в окрестности, чтобы точка считалась центральной(параметр HDBSCAN).
        cluster_selection_epsilon: float, default=0.0
            Максимальное расстояние, допустимое между точками, чтобы они считались связанными в процессе кластеризации на основе плотности(параметр HDBSCAN).

        Returns
        -------
        np.ndarray
            Массив меток размера N.
        """

        assert data.ndim == 2, "data: Ожидался 2D массив"
        cluster = HDBSCANClassifier.get_cluster(data, min_cluster_size, min_samples, cluster_selection_epsilon)
        labels = HDBSCANClassifier.__get_labels(cluster)
        return cluster, labels
